
/*
*****************************************************************************************************
*****************************************************************************************************
*****                                                                                           *****
*****                  Task D   -   Area Fill Routine                                           *****
*****                                                                                           *****
*****      Fill routine is adapted from -    TINT FILL by Alvy Ray Smith                        *****
*****                                           Computer Graphics Lab                           *****
*****                                       New York Institute of Technology                    *****
*****                                          Old Westbury, NY  11568                          *****
*****                                                                                           *****
*****                                                                                           *****
*****   The area to fill is specified by an X & Y start point, fill color and boundary color    *****
*****                                                                                           *****
*****************************************************************************************************
*****************************************************************************************************
*/

#include	"styx.h"
#define		STACK_MAX	1000
#define		X_VAL		0
#define		Y_VAL		1
#define		T_LIM		99
#define		B_LIM		0
#define		L_LIM		0
#define		R_LIM		159


int	f_x, f_y, f_lx, f_rx;
int	xref, yref, lxref, rxref;
int	new_color, old_color, new_mask, old_mask, boundary_mask, f_colision;
int	savex_x, savexy_x, savexy_y;	/* teperaries used by savex() & savexy() */
int	stack_pointer, stack_big;
int	stack[2][STACK_MAX];



/*	
	name		polyfill	Fills an area within a poligon border, border points are NOT specified


	synopsis        polyfill( x, y, fill_color, boundary_color);
			int	x, y, fill_color, boundary_color;

	discription	This fill utility works with Low-Res graphics to fill an area 
			from the start point to the boundary line with the fill_color.
			The value returned represents a colision detection of any colors 
			written over during the fill operation.

			This vertion has special features for styx:
				- a border color of grey is used for a secondary border
				- if a fill is attempted over dark or light white then it is aborted 
*/


polyfill(seedx,seedy,f_color,b_color)
int	seedx, seedy, f_color, b_color;
{

	/* ================ Main body of fill routine =============== */

	stack_pointer=0; stack_big=0;		/* initialize stack */
	f_x=seedx; f_y=seedy;			/* initialize start point of fill */
	boundary_mask=get_mask(b_color);	/* get low-res mask for boundary color */
	new_color=f_color;			/* get new_color for all routines to use */
	new_mask=get_mask(f_color);		/* get low-res mask for new color */
	old_mask=xpset(f_x,f_y,BLACK);		/* find old background color */
	old_color=get_color(old_mask);
	f_colision=0;				/* clear colision detector */

/*	cprintf(" old color %d \n\r",old_mask);  */

	if(old_mask==new_mask || (old_mask & 0x8080) ) {	/* if area is already filled then exit */
#ifdef	DEBUG
	cprintf("area already filled or white   \n\r ");
#endif
		return(TRUE);
	}
	yref=f_y;	push();		/* mark first place to begin fill */

	while( stack_not_empty() ) {	/* fill untill done */ 
#ifdef	DEBUG
		exit_check();		
#endif
		pop();
		if(xpset(f_x,f_y,BLACK)==new_mask)
			continue;
		filline();
		scanshore();
		if(hineighbor())
			scanhi();
		else {
			if(loneighbor() )
				scanlo();
			else {
				scanhi(); scanlo();
			}
		}

		yref=f_y; lxref=f_lx; rxref=f_rx;

	}
#ifdef	DEBUG
	if(stack_big>150)
		cprintf("stack size grew to  %d  \n\r ",stack_big);
#endif
	return(f_colision);	/* return colision info */

}




filline()
{
	fillright(); filleft();

}



fillright()
{
	register	temp, flag;

	savex(); 
	
/*
	flag=TRUE;
	while(flag) {
		temp=pset(f_x,f_y,new_color);
		if(temp!=boundary_mask && temp!=0x0100 && temp!=new_mask && f_x<R_LIM) {
			f_x++;
			f_colision|=temp;
			fill_count++;
		}
		else {
			pset(f_x,f_y,get_color(temp) );
			flag=FALSE;
		}
	}
*/

	f_x=fright(f_x,f_y,new_color);
		
	f_rx=f_x-1; restx();

}


filleft()
{
        register	temp, flag;
	
	savex(); f_x--;
/*
	flag=TRUE;
	while(flag) {
		temp=pset(f_x,f_y,new_color);
		if(temp!=boundary_mask && temp!=0x0100 && temp!=new_mask && f_x>L_LIM) {
			f_x--;
			f_colision|=temp;
			fill_count++;
		}
		else {
			pset(f_x,f_y,get_color(temp) );
			flag=FALSE;
		}
	}
*/


	f_x=fleft(f_x,f_y,new_color);

	f_lx=f_x+1; restx();
}



scanshort()
{

	savexy(); f_x=f_lx; f_y++;

	if(f_x==f_rx && xpset(f_x,f_y,BLACK)==old_mask)
		push();

	f_y=f_y-2;

	if(f_x==f_rx && xpset(f_x,f_y,BLACK)==old_mask)
		push();

	restxy();
}



scanhi()
{
	if(f_y+1>T_LIM)
		return(TRUE);

	savexy(); f_x=f_lx; f_y++;

	while(f_x<f_rx) {

/*
		while(xpset(f_x,f_y,BLACK)!=old_mask  && f_x<f_rx)
			f_x++;
*/
		f_x=scanne(f_x,f_y,old_color,f_rx);

		if(f_x>f_rx)
			break;
		push();

/*
		while(xpset(f_x,f_y,BLACK)==old_mask && f_x<f_rx) 
			f_x++;
*/

		f_x=scaneq(f_x,f_y,old_color,f_rx);

	}


	restxy();

}


scanlo()
{
	if(f_y-1<B_LIM)
		return(TRUE);

	savexy(); f_x=f_lx; f_y--;

	while(f_x<f_rx) {


/*
		while(xpset(f_x,f_y,BLACK)!=old_mask  && f_x<f_rx)
			f_x++;
*/
		f_x=scanne(f_x,f_y,old_color,f_rx);

		if(f_x>f_rx)
			break;
		push();


/*
		while(xpset(f_x,f_y,BLACK)==old_mask && f_x<f_rx) 
			f_x++;
*/

		f_x=scaneq(f_x,f_y,old_color,f_rx);

	}

	restxy();

}


	
hineighbor()
{
	if(f_y==yref+1 && f_lx>lxref-1 && f_rx<rxref+1)
		return(TRUE);
	else
		return(FALSE);
}


loneighbor()
{
	if(f_y==yref-1 && f_lx>lxref-1 && f_rx<rxref+1)
		return(TRUE);
	else
		return(FALSE);
}



push()
{
	stack[X_VAL][stack_pointer]=f_x;	
	stack[Y_VAL][stack_pointer]=f_y;	
	stack_pointer++;
	if(stack_pointer>stack_big)
		stack_big=stack_pointer;

	if(stack_pointer>=STACK_MAX) {

#ifdef	DEBUG
		cprintf("stack overflow \n\r ");

#endif
		stack_pointer--;
	}

}


pop()
{
	stack_pointer--;
	if(stack_pointer<0)
		stack_pointer=0;
	f_x=stack[X_VAL][stack_pointer];
	f_y=stack[Y_VAL][stack_pointer];

}


stack_not_empty()
{
	if(stack_pointer>0)
		return(TRUE);	/* stack is NOT empty */
	else
		return(FALSE);

}


savex()
{
	savex_x=f_x;
}



restx()
{
	f_x=savex_x;
}



savexy()
{
	savexy_x=f_x;
	savexy_y=f_y;
}


restxy()
{
	f_x=savexy_x;
	f_y=savexy_y;
}


get_mask(temp)
int	temp;
{
	switch(temp) {
	case BLACK:
		return(0x0001);
	case DARK_BLUE:
		return(0x0002);
	case DARK_GREEN:
		return(0x0004);
	case DARK_CYAN:
		return(0x0008);
	case DARK_RED:
		return(0x0010);
	case DARK_MAGENTA:
		return(0x0020);
	case DARK_BROWN:
		return(0x0040);
	case DARK_WHITE:
		return(0x0080);
	case GREY:
		return(0x0100);
	case LIGHT_BLUE:
		return(0x0200);
	case LIGHT_GREEN:
		return(0x0400);
	case LIGHT_CYAN:
		return(0x0800);
	case LIGHT_RED:
		return(0x1000);
	case LIGHT_MAGENTA:
		return(0x2000);
	case YELLOW:
		return(0x4000);
	case LIGHT_WHITE:
		return(0x8000);
	default:
		break;
	}
}


get_color(mask)
int	mask;
{
	switch(mask) {
	case 0x0001:
		return(BLACK);
	case 0x0002:
		return(DARK_BLUE);
	case 0x0004:
		return(DARK_GREEN);
	case 0x0008:
		return(DARK_CYAN);
	case 0x0010:
		return(DARK_RED);
	case 0x0020:
		return(DARK_MAGENTA);
	case 0x0040:
		return(DARK_BROWN);
	case 0x0080:
		return(DARK_WHITE);
	case 0x0100:
		return(GREY);
	case 0x0200:
		return(LIGHT_BLUE);
	case 0x0400:
		return(LIGHT_GREEN);
	case 0x0800:
		return(LIGHT_CYAN);
	case 0x1000:
		return(LIGHT_RED);
	case 0x2000:
		return(LIGHT_MAGENTA);
	case 0x4000:
		return(YELLOW);
	case 0x8000:
		return(LIGHT_WHITE);
	default:
		break;
	}
}
